// PilBoxSim.cpp: implementation of the CPilBoxSim class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PilBoxSim.h"
#include "Interface.h"

// PILBox commands
#define TDIS		0x494					// TDIS
#define COFI		0x495					// COFF with IDY, firmware >= v1.6
#define CON			0x496					// Controller ON
#define COFF		0x497					// Controller OFF
#define SSRQ		0x49C					// Set Service Request
#define CSRQ		0x49D					// Clear Service Request

// IL flags
#define IL_FRREC	(1 << 0)				// frame received
#define IL_FRAV		(1 << 1)				// frame available  
#define IL_FCMD		(1 << 2)				// command frame received
#define IL_SSRQ		(1 << 3)				// set SRQ bit in IDY
#define IL_MBOX		(1 << 5)				// mailbox mode
#define IL_TRENA	(1 << 6)				// translator enable 
#define IL_CTRL		(1 << 7)				// controller on (in translator mode)

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPilBoxSim::CPilBoxSim(bool bEnableRFC)
	: m_bEnableRFC(bEnableRFC)
	, m_bUSE8BITS(true)
	, m_byLastHigh(0)
	, m_byLastByte(0)
	, m_byIlFlg(0)
{
	InitializeCriticalSection(&m_csLock);
}

CPilBoxSim::~CPilBoxSim()
{
	DeleteCriticalSection(&m_csLock);
}

//
// set new loop callback function
//
void CPilBoxSim::SetLoopCallback(CIL* pIL)
{
	EnterCriticalSection(&m_csLock);
	{
		m_pIL = pIL;
	}
	LeaveCriticalSection(&m_csLock);
	return;
}

//
// PIL-Box byte handler
//
BYTE CPilBoxSim::PILBox(BYTE byVal)
{
	WORD wFrame;							// IL frame
	BYTE byData;							// answer

	TRACE(_T("PILBox: %02X\n"),byVal);

	if ((byVal & 0xE0) == 0x20)				// fetched high byte
	{
		//TRACE(_T("PILBox High Byte: %02X\n"),byVal);

		m_byLastHigh = byVal;				// high byte, save it
		byData = 0;							// no return
	}
	else if ((byVal & 0xC0) != 0)			// fetched low byte
	{
		//TRACE(_T("PILBox Low  Byte: %02X\n"),byVal);

		// low byte, build frame according to format
		if (m_bUSE8BITS = ((byVal & 0x80) != 0))
			wFrame = ((m_byLastHigh & 0x1E) << 6) | (byVal & 0x7F);
		else
			wFrame = ((m_byLastHigh & 0x1F) << 6) | (byVal & 0x3F);

		//
		// frame handling
		//
		if ((wFrame & 0x7F4) != TDIS)		// no special PIL-Box cmd
		{
			if ((m_byIlFlg & IL_TRENA))		// translator enabled (don't care about mode, only CON mode simulated)
			{
				EnterCriticalSection(&m_csLock);
				{
					if (m_pIL != nullptr)	// has external loop
					{
						// transmit IL frame to internal virtual devices
						TRACE(_T("PILBox frame data to virtual HP-IL loop : %02X\n"),wFrame);

						ASSERT(m_pIL);
						wFrame = m_pIL->Device(wFrame);

						// RFC frame enabled and received a CMD frame from the controller
						if (m_bEnableRFC && (wFrame & 0x700) == 0x400)
						{
							m_pIL->Device(0x500);
						}
					}
				}
				LeaveCriticalSection(&m_csLock);
			}

			// build the low and high bytes for the answer
			if (m_bUSE8BITS == false)
			{
				// use 7-bit characters for maximum compatibility
				m_byTxHigh = ((wFrame >> 6) & 0x1F) | 0x20;
				m_byTxLow  = (wFrame & 0x3F) | 0x40;
			}
			else
			{
				// 8-bit format for optimum speed
				m_byTxHigh = ((wFrame >> 6) & 0x1E) | 0x20;
				m_byTxLow  = (wFrame & 0x7F) | 0x80;
			}

			if (m_byTxHigh != m_byLastHigh)
			{
				// send high part if different from last frame
				m_byLastHigh = m_byTxHigh;
				byData = m_byTxHigh;
			}
			else
			{
				// otherwise send only low part
				byData = m_byTxLow;
			}
		}
		else								// special PIL-Box cmd
		{
			TRACE(_T("PILBox command : %02X\n"),wFrame);

			// decode PILBox command
			if (wFrame == TDIS)
			{
				m_byIlFlg &= ~IL_TRENA;		// disable translator enable 
			}
			else
			{
				m_byIlFlg |= IL_TRENA;		// enable translator enable 
			}

			byData = byVal;					// just return the low-byte in the original format
		}
	}
	else
	{
		// ack for last high byte transfer
		if (byVal == 13 && (m_byLastByte & 0xE0) == 0x20)
		{
			//TRACE(_T("PILBox Ack  Byte: %02X\n"),byVal);
			byData = m_byTxLow;				// the low byte on ack
		}
		else
		{
			byData = 0;						// no answer
		}
	}
	m_byLastByte = byData;
	return byData;
}
