// PipeServ.cpp: implementation of the CPipeServ class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Ilpilsim.h"
#include "PipeServ.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define CONNECTING_STATE 0
#define READING_STATE 1
#define WRITING_STATE 2
#define PIPE_TIMEOUT 5000

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPipeServ::CPipeServ(CIL *pIL, LPCTSTR lpszPipeName, CWnd* pParent)
	: m_strPipeName(lpszPipeName)
	, CPilBoxSim(true)
{
	m_pIL = pIL;
	m_pParent = pParent;
}

CPipeServ::~CPipeServ()
{
}

// --------------------------------
// StartServer()
// start the local named pipe server
// --------------------------------

bool CPipeServ::StartServer()
{
	// Create an instances of a named pipe along with an event
	// object. An overlapped ConnectNamedPipe operation is
	// started for this instance.
	ZeroMemory(&m_Pipe,sizeof(m_Pipe));

	// Create an event object for this instance
	m_Pipe.oOverlap.hEvent = CreateEvent(
		NULL,								// default security attribute
		TRUE,								// manual-reset event
		TRUE,								// initial state = signaled
		NULL);								// unnamed event object

	if (m_Pipe.oOverlap.hEvent == nullptr)
	{
		return false;
	}

	m_Pipe.hPipe = CreateNamedPipe(
		m_strPipeName,						// pipe name
		PIPE_ACCESS_DUPLEX |				// read/write access
		FILE_FLAG_OVERLAPPED,				// overlapped mode
		PIPE_TYPE_BYTE |					// message-type pipe
		PIPE_READMODE_BYTE |				// message-read mode
		PIPE_WAIT,							// blocking mode
		1,									// number of instances
		1,									// output buffer size
		1,									// input buffer size
		PIPE_TIMEOUT,						// client time-out
		NULL);								// default security attributes

	if (m_Pipe.hPipe == INVALID_HANDLE_VALUE)
	{
		CloseHandle(m_Pipe.oOverlap.hEvent);
		return false;
	}

	// Call the subroutine to connect to the new client
	m_Pipe.fPendingIO = ConnectToNewClient(m_Pipe.hPipe,&m_Pipe.oOverlap);

	m_Pipe.dwState = m_Pipe.fPendingIO
				   ? CONNECTING_STATE		// still connecting
				   : READING_STATE;			// ready to read

	m_bRunning = true;
	AfxBeginThread(ThreadPipeServer,this);
	return true;
}

// --------------------------------
// StopServer()
// stop the local TCP/IP server
// --------------------------------

void CPipeServ::StopServer()
{
	if (m_bRunning)
	{
		m_bRunning = false; 				// delete server thread
		SetEvent(m_Pipe.oOverlap.hEvent);
		CloseHandle(m_Pipe.hPipe);
		CloseHandle(m_Pipe.oOverlap.hEvent);
	}
	return;
}

// --------------------------------
// ThreadPipeServer
// local pipe server thread
// --------------------------------

UINT CPipeServ::ThreadPipeServer(LPVOID pParam)
{
	CPipeServ *p = reinterpret_cast<CPipeServ *>(pParam);

	DWORD cbRet,dwErr;
	BOOL  fSuccess;

	while (p->m_bRunning)
	{
		// Wait for the event object to be signaled, indicating
		// completion of an overlapped read, write, or
		// connect operation
		WaitForSingleObject(
			p->m_Pipe.oOverlap.hEvent,		// event object
			INFINITE);						// waits indefinitely

		// Get the result if the operation was pending
		if (p->m_Pipe.fPendingIO)
		{
			fSuccess = GetOverlappedResult(
				p->m_Pipe.hPipe,			// handle to pipe
				&p->m_Pipe.oOverlap,		// OVERLAPPED 
				&cbRet,						// bytes transferred
				FALSE);						// do not wait

			switch (p->m_Pipe.dwState)
			{
			// Pending connect operation
			case CONNECTING_STATE:
				if (!fSuccess)
				{
					continue;
				}
				p->m_Pipe.dwState = READING_STATE;
				break;
			// Pending read operation
			case READING_STATE:
				if (!fSuccess || cbRet == 0)
				{
					p->DisconnectAndReconnect();
					continue;
				}
				p->m_Pipe.cbRead = cbRet;
				p->m_Pipe.dwState = WRITING_STATE;
				break;
			// Pending write operation
			case WRITING_STATE:
				if (!fSuccess || cbRet != p->m_Pipe.cbToWrite)
				{
					p->DisconnectAndReconnect();
					continue;
				}
				p->m_Pipe.dwState = READING_STATE;
				break;
			default:
				continue;
			}
		}

		// The pipe state determines which operation to do next
		switch (p->m_Pipe.dwState)
		{
		// READING_STATE:
		// The pipe instance is connected to the client
		// and is ready to read a request from the client
		case READING_STATE:
			fSuccess = ReadFile(
				p->m_Pipe.hPipe,
				&p->m_Pipe.byRequest,
				sizeof(p->m_Pipe.byRequest),
				&p->m_Pipe.cbRead,
				&p->m_Pipe.oOverlap);

			// The read operation completed successfully
			if (fSuccess && p->m_Pipe.cbRead != 0)
			{
				p->m_Pipe.fPendingIO = false;
				p->m_Pipe.dwState = WRITING_STATE;
				continue;
			}

			// The read operation is still pending
			dwErr = GetLastError();
			if (!fSuccess && (dwErr == ERROR_IO_PENDING))
			{
				p->m_Pipe.fPendingIO = true;
				continue;
			}

			// An error occurred; disconnect from the client
			p->DisconnectAndReconnect();
			break;

		// WRITING_STATE:
		// The request was successfully read from the client.
		// Get the reply data and write it to the client.
		case WRITING_STATE:
			p->m_Pipe.byReply = p->PILBox(p->m_Pipe.byRequest);	
			if (p->m_Pipe.byReply != 0)		// answer
			{
				// transmit
				p->m_Pipe.cbToWrite = 1;

				fSuccess = WriteFile(
					p->m_Pipe.hPipe,
					&p->m_Pipe.byReply,
					p->m_Pipe.cbToWrite,
					&cbRet,
					&p->m_Pipe.oOverlap);

				// The write operation completed successfully
				if (fSuccess && cbRet == p->m_Pipe.cbToWrite)
				{
					p->m_Pipe.fPendingIO = false;
					p->m_Pipe.dwState = READING_STATE;
					continue;
				}

				// The write operation is still pending
				dwErr = GetLastError();
				if (!fSuccess && (dwErr == ERROR_IO_PENDING))
				{
					p->m_Pipe.fPendingIO = true;
					continue;
				}

				// An error occurred; disconnect from the client
				p->DisconnectAndReconnect();
			}
			else // no answer
			{
				p->m_Pipe.fPendingIO = false;
				p->m_Pipe.dwState = READING_STATE;
				continue;
			}
			break;
		default:
			continue;
		}
	}

	p->m_pParent->PostMessage(WM_SERVERSTOPPED,reinterpret_cast<WPARAM>(p));
	AfxEndThread(0);
	return 0;
}

// DisconnectAndReconnect()
// This function is called when an error occurs or when the client
// closes its handle to the pipe. Disconnect from this client, then
// call ConnectNamedPipe to wait for another client to connect.

void CPipeServ::DisconnectAndReconnect()
{
	// Disconnect the pipe instance
	DisconnectNamedPipe(m_Pipe.hPipe);

	// Call a subroutine to connect to the new client
	m_Pipe.fPendingIO = ConnectToNewClient(m_Pipe.hPipe,&m_Pipe.oOverlap);
	
	m_Pipe.dwState = m_Pipe.fPendingIO
				   ? CONNECTING_STATE		// still connecting
				   : READING_STATE;			// ready to read
	return;
}

// ConnectToNewClient(HANDLE, LPOVERLAPPED)
// This function is called to start an overlapped connect operation.
// It returns TRUE if an operation is pending or FALSE if the
// connection has been completed.

bool CPipeServ::ConnectToNewClient(HANDLE hPipe, LPOVERLAPPED lpo)
{
	BOOL fConnected;
	bool fPendingIO = false;

	// Start an overlapped connection for this pipe instance
	fConnected = ConnectNamedPipe(hPipe, lpo);

	// Overlapped ConnectNamedPipe should return zero
	if (fConnected)
	{
		return false;
	}

	switch (GetLastError())
	{
	// The overlapped connection in progress
	case ERROR_IO_PENDING:
		fPendingIO = true;
		break;

	// Client is already connected, so signal an event
	case ERROR_PIPE_CONNECTED:
		if (SetEvent(lpo->hEvent))
			break;

	// If an error occurs during the connect operation...
	default:
		return false;
	}
	return fPendingIO;
}
